from herre.herre import Herre
from fakts.grants.remote import Manifest
from fakts import Fakts
from herre.grants import CacheGrant
from herre.grants.oauth2.refresh import RefreshGrant
from herre.grants.fakts import FaktsGrant
from typing import Optional
from herre.grants.fakts.fakts_login_screen import FaktsQtLoginScreen
from herre.grants.qt.login_screen import LoginWidget


class ArkitektHerreRefreshGrant(RefreshGrant):
    grant: FaktsGrant


class ArkitektHerreChooseFaktsQt(FaktsQtLoginScreen):
    grant: RefreshGrant


class ArkitektHerreQtCacheGrant(CacheGrant):
    grant: ArkitektHerreChooseFaktsQt


class ArkitektHerreQt(Herre):
    grant: ArkitektHerreQtCacheGrant


def build_arkitekt_qt_herre(
    manifest: Manifest,
    fakts: Fakts,
    no_cache: Optional[bool] = False,
    login_widget=None,
    parent=None,
):
    identifier = manifest.identifier
    version = manifest.version

    login_widget = login_widget or LoginWidget(identifier, version, parent=parent)

    return ArkitektHerreQt(
        grant=ArkitektHerreQtCacheGrant(
            cache_file=f".arkitekt/cache/{identifier}-{version}_herre_cache.json",
            hash=f"{identifier}-{version}",
            skip_cache=not no_cache,
            grant=ArkitektHerreChooseFaktsQt(
                fakts=fakts,
                widget=login_widget,
                auto_login=True,
                grant=ArkitektHerreRefreshGrant(
                    grant=FaktsGrant(fakts=fakts, fakts_group="lok")
                ),
            ),
        ),
    )



from typing import List
from rekuest.api.schema import (
    NodeFragment,
    ArgPortFragment,
    ChildPortFragment,
    PortKind,
    ReturnPortFragment,
    retrieveall
)
    

import ast
import re
import inflection
from arkitekt import easy

def clean(varStr):
    return re.sub("\W|^(?=\d)", "_", varStr)


identifier_class_map = {
    "representation": "mikro.Representation",
    "omerofile": "mikro.OmeroFile",
    "sample": "mikro.Sample",
    "imageplus": "mikro.ImagePlus",
    "metric": "mikro.Metric",
    "thumbnail": "mikro.Thumbnail",
}

shrinker_map = {
    "@mrepresentation": "Representation",
    "omerofile": "OmeroFile",
    "sample": "Sample",
    "imageplus": "ImagePlus",
    "metric": "Metric",
    "thumbnail": "Thumbnail",
}

expander_map = shrinker_map


def generate_imports():

    imports = []

    for key, item in identifier_class_map.items():
        imports.append(
            ast.ImportFrom(
                module=".".join(item.split(".")[:-1]),
                names=[ast.alias(name=item.split(".")[-1])],
                level=0,
            )
        )

    imports += [
        ast.ImportFrom(
            module="enum",
            names=[ast.alias(name="Enum")],
            level=0,
        ),
        ast.ImportFrom(
            module="pydantic.fields",
            names=[ast.alias(name="Field")],
            level=0,
        ),
        ast.ImportFrom(
            module="typing",
            names=[
                ast.alias(name="Optional"),
                ast.alias(name="List"),
                ast.alias(name="Dict"),
                ast.alias(name="Union"),
                ast.alias(name="Literal"),
            ],
            level=0,
        ),
    ]

    return imports

def wrap_nullable(arg: ArgPortFragment | ChildPortFragment | ReturnPortFragment, x):
    if arg.nullable:
        return ast.Subscript(
                    value=ast.Name("Optional", ctx=ast.Load()),
                    slice=x,
                )

    return x

def get_arg_annotation(arg: ArgPortFragment | ChildPortFragment | ReturnPortFragment):
    if arg.kind == PortKind.STRUCTURE:
        return wrap_nullable(arg, ast.Name(
                        id = "ID",
                        ctx=ast.Load()))

    if arg.kind == PortKind.INT:
        return wrap_nullable(arg, ast.Name(
                    id="int",
                    ctx=ast.Load(),
                ))
    if arg.kind == PortKind.FLOAT:
        return wrap_nullable(arg, ast.Name(
                    id="float",
                    ctx=ast.Load(),
                ))
    if arg.kind == PortKind.STRING:
        return wrap_nullable(arg, ast.Name(
                    id="str",
                    ctx=ast.Load(),
                ))
    if arg.kind == PortKind.BOOL:
        return wrap_nullable(arg, ast.Name(
                    id="bool",
                    ctx=ast.Load()))
    if arg.kind == PortKind.LIST:
        return wrap_nullable(arg, ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()),
                    slice=get_arg_annotation(arg.child),
                ))
    if arg.kind == PortKind.DICT:
        return wrap_nullable(arg, ast.Subscript(
                    value=ast.Name("Dict", ctx=ast.Load()),
                    slice=ast.Tuple(elts=[ast.Name("Str", ctx=ast.Load()), get_arg_annotation(arg.child)]),
                ))
    


def generate_node_args(
    args: List[ArgPortFragment]
):

    pos_args = []
    kw_values = []

    for arg in args:
        get_arg_annotation(arg)
        if arg.default:
            kw_values.append(ast.Constant(value=arg.default))

    return ast.arguments(
        args=pos_args,
        posonlyargs=[],
        kwonlyargs=[],
        kw_defaults=[],
        defaults=kw_values,
    )


def generate_node_returns(returns: List[ReturnPortFragment]):

    if len(returns) == 0:
        return ast.Name(id="None")

    if len(returns) == 1:
        return get_arg_annotation(returns[0])
    
    return ast.Tuple(
        elts=[get_arg_annotation(re) for re in returns],
        ctx=ast.Load(),
    )



def generate_node_doc(node: NodeFragment):

    description = f"{node.name}\n\n{node.description}\n"

    if len(node.args) > 0:
        description += "\nArguments:\n"
        for arg in node.args:
            description += f"\t{arg.key} ({arg.kind}) {arg.description}\n"

    if len(node.returns) > 0:
        description += "\nReturns:\n"
        for re in node.returns:
            description += f"\t{re.key} ({re.kind}) {re.description}\n"

    return ast.Expr(value=ast.Constant(value=description))


def node_to_ast(node: NodeFragment):

    return [
        ast.AsyncFunctionDef(
            name=f"{inflection.underscore(node.name.replace(' ', '_'))}",
            args=generate_node_args(node.args),
            body=[
                generate_node_doc(node),
                ast.Return(value=ast.Constant(value=None)),
            ],
            decorator_list=[],
            returns=generate_node_returns(node.returns),
        )
    ]


def generate_node_func(
    file="generated_node.py"
):

    global_tree = []

    global_tree = generate_imports()

    with easy("nodifier"):
        for node in retrieveall():
            global_tree += node_to_ast(node)

    md = ast.Module(body=global_tree, type_ignores=[])
    generated = ast.unparse(ast.fix_missing_locations(md))

    #res = format_str(generated, mode=FileMode())

    with open(file, "w") as f:
        f.write(generated)


if __name__ == "__main__":
    generate_node_func()






from typing import List, Optional
import logging
import os
from .utils import create_arkitekt_folder
from fakts.grants.remote import Manifest


def easy(
    identifier: str,
    version: str = "latest",
    logo: str = None,
    scopes: List[str] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    allow_sync_in_async: bool = True,
    log_level: str = "ERROR",
    token: str = None,
    no_cache: bool = False,
    instance_id: str = "main",
):
    """Easy app creation

    A simple way to create an Arkitekt app with a device code grant
    it will cache the configuration in a file called `identifier`_token.json
    and the cache in a file called `identifier`_cache.json

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        url (_type_, optional): The app configuration url. Defaults to "http://localhost:8000/f/".
        headless (bool, optional): Do not open a browser window. Defaults to False.
        allow_sync_in_async (bool, optional): Should we allow the creation of a sync interface in an async loop (necessary for a sync interface in jupyter). Defaults to True.

    Returns:
        Arkitekt: _description_
    """
    from arkitekt.apps.default import App, build_arkitekt_app

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    create_arkitekt_folder(with_cache=True)

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_app(
        manifest=manifest,
        url=url,
        no_cache=no_cache,
        headless=headless,
        instance_id=instance_id,
    )

    try:
        from reaktion.extension import ReaktionExtension

        app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()
    except ImportError:
        pass

    return app


def next(
    identifier: str,
    version: str = "latest",
    logo: str = None,
    scopes: List[str] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    allow_sync_in_async: bool = True,
    log_level: str = "ERROR",
    token: str = None,
    no_cache: bool = False,
    instance_id: str = "main",
):
    """Next app creation

    A simple way to create an Arkitekt app with a device code grant and
    all new features enabled. This will currently replace the
    mikro service with a new one that supports new features, requires
    the mikro_next package to be installed.


    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        url (_type_, optional): The app configuration url. Defaults to "http://localhost:8000/f/".
        headless (bool, optional): Do not open a browser window. Defaults to False.
        allow_sync_in_async (bool, optional): Should we allow the creation of a sync interface in an async loop (necessary for a sync interface in jupyter). Defaults to True.

    Returns:
        Arkitekt: _description_
    """
    from arkitekt.apps.default import App, build_arkitekt_app

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    create_arkitekt_folder(with_cache=True)

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_app(
        manifest=manifest,
        url=url,
        no_cache=no_cache,
        headless=headless,
        instance_id=instance_id,
    )

    return app


def jupy(
    identifier: str,
    version: str = "latest",
    url: str = "http://localhost:8000",
    headless: bool = False,
    instance_id: str = "main",
    no_cache: bool = False,
):
    app = easy(
        identifier,
        version,
        url=url,
        headless=headless,
        allow_sync_in_async=True,
        log_level="ERROR",
        instance_id=instance_id,
        no_cache=no_cache,
    )
    app.koil.sync_in_async = True
    app.enter()  # This will start the event loop with the sync interface (for automatic awaits in jupyter)
    return app


def port(
    identifier: str,
    version: str = "latest",
    url: str = "http://lok:8000",
    log_level: str = "ERROR",
    token: str = None,
    headless: bool = False,
    instance_id: str = "main",
    no_cache: bool = False,
):
    """Easy port creation

    A simple way to create an Arkitekt app with a device code grant
    it will cache the configuration in a file called `identifier`_token.json
    and the cache in a file called `identifier`_cache.json

    Args:
        identifier (str): The apps identifier
        version (str, optiosnal): The apps verion. Defaults to "latest".
        url (_type_, optional): The app configuration url. Defaults to "http://localhost:8000/f/".
        headless (bool, optional): Do not open a browser window. Defaults to False.
        allow_sync_in_async (bool, optional): Should we allow the creation of a sync interface in an async loop (necessary for a sync interface in jupyter). Defaults to True.

    Returns:
        Arkitekt: _description_
    """
    from arkitekt.apps.default import App, build_arkitekt_app

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    if not token:
        raise ValueError("You must provide a token")

    if not url:
        raise ValueError("You must provide a url")

    create_arkitekt_folder(with_cache=True)

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    raise NotImplementedError("Port is not implemented yet")


def scheduler(
    identifier: str,
    version: str = "latest",
    logo: str = None,
    scopes: List[str] = None,
    url: str = "http://localhost:8000",
    headless: bool = False,
    allow_sync_in_async: bool = True,
    log_level: str = "INFO",
    token: str = None,
    instance_id: str = "main",
    no_cache: bool = False,
):
    """Scheduler app creation

    A an arkitekt scheduler app with a device code grant and a reaktion agent
    this app will be able to deploy graphs from fluss

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        url (_type_, optional): The app configuration url. Defaults to "http://localhost:8000/f/".
        headless (bool, optional): Do not open a browser window. Defaults to False.
        allow_sync_in_async (bool, optional): Should we allow the creation of a sync interface in an async loop (necessary for a sync interface in jupyter). Defaults to True.

    Returns:
        Arkitekt: _description_
    """
    from arkitekt.apps.default import App, build_arkitekt_app
    from arkitekt.apps.rekuest import ArkitektWebsocketAgentTransport

    try:
        from reaktion.extension import ReaktionExtension
    except ImportError as e:
        raise ImportError(
            "You need to install reaktion to use the scheduler function"
        ) from e

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    create_arkitekt_folder(with_cache=True)

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_app(
        manifest=manifest,
        url=url,
        no_cache=no_cache,
        headless=headless,
        instance_id=instance_id,
    )
    app.rekuest.agent.extensions["reaktion"] = ReaktionExtension()

    return app


def publicqt(
    identifier: str,
    version: str = "latest",
    parent=None,
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    beacon_widget=None,
    login_widget=None,
    force_herre_grant=None,
    instance_id: str = "main",
    no_cache: bool = False,
    log_level: str = "ERROR",
):
    """Public QtApp creation

    A simple way to create an Arkitekt app with a public grant (allowing users to sign
    in with the application ) utlizing a retrieve grant (necessating a previous configuration
    of the application on the server side)

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        parent (QtWidget, optional): The QtParent (for the login and server select widget). Defaults to None.

    Returns:
        Arkitekt: The Arkitekt app
    """

    from arkitekt.apps.qt import build_arkitekt_qt_app

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes or ["openid", "read", "write"],
        logo=logo,
    )

    create_arkitekt_folder(with_cache=True)

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = build_arkitekt_qt_app(
        manifest=manifest,
        parent=parent,
        no_cache=no_cache,
        beacon_widget=beacon_widget,
        login_widget=login_widget,
        instance_id=instance_id,
    )
    app.enter()  # This will start the event loop and attach the event handlers
    return app


def publicscheduleqt(
    identifier: str,
    version: str = "latest",
    parent=None,
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    beacon_widget=None,
    login_widget=None,
    force_herre_grant=None,
    instance_id: str = "main",
    no_cache: bool = False,
    log_level: str = "ERROR",
):
    """Public QtApp creation

    A simple way to create an Arkitekt app with a public grant (allowing users to sign
    in with the application ) utlizing a retrieve grant (necessating a previous configuration
    of the application on the server side)

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        parent (QtWidget, optional): The QtParent (for the login and server select widget). Defaults to None.

    Returns:
        Arkitekt: The Arkitekt app
    """

    from arkitekt.apps.scheduleqt import build_arkitekt_scheduleqt_app

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes or ["openid", "read", "write"],
        logo=logo,
    )

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    create_arkitekt_folder(with_cache=True)

    x = build_arkitekt_scheduleqt_app(
        manifest=manifest,
        parent=parent,
        no_cache=no_cache,
        beacon_widget=beacon_widget,
        login_widget=login_widget,
        instance_id=instance_id,
    )
    x.enter()
    return x



from arkitekt.healthz import FaktsHealthz
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.links.split import SplitLink
from rath.contrib.fakts.links.subscription_transport_ws import FaktsWebsocketLink
from rath.contrib.herre.links.auth import HerreAuthLink
from rekuest.rath import RekuestLinkComposition, RekuestRath
from rekuest.rekuest import Rekuest
from graphql import OperationType
from rekuest.contrib.arkitekt.websocket_agent_transport import (
    ArkitektWebsocketAgentTransport,
)
from rekuest.agents.base import BaseAgent
from fakts import Fakts
from herre import Herre
from rekuest.postmans.graphql import GraphQLPostman


class ArkitektRekuest(Rekuest):
    rath: RekuestRath
    agent: BaseAgent
    healthz: FaktsHealthz


def build_arkitekt_rekuest(
    fakts: Fakts, herre: Herre, instance_id: str
) -> ArkitektRekuest:
    rath = RekuestRath(
        link=RekuestLinkComposition(
            auth=HerreAuthLink(herre=herre),
            split=SplitLink(
                left=FaktsAIOHttpLink(fakts_group="rekuest", fakts=fakts),
                right=FaktsWebsocketLink(fakts_group="rekuest", fakts=fakts),
                split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
            ),
        )
    )

    return ArkitektRekuest(
        rath=rath,
        agent=BaseAgent(
            transport=ArkitektWebsocketAgentTransport(
                fakts_group="rekuest.agent", fakts=fakts, herre=herre
            ),
            instance_id=instance_id,
            rath=rath,
        ),
        postman=GraphQLPostman(rath=rath),
        healthz=FaktsHealthz(fakts_group="rekuest", fakts=fakts),
    )



from fakts.fakts import Fakts
from koil.composition.base import Composition
from pydantic import Field
from fakts.fakts import Fakts
from fakts.grants.remote.static import StaticGrant
from fakts.grants import CacheGrant
from herre.grants import CacheGrant as HerreCacheGrant
from herre.grants.oauth2.refresh import RefreshGrant
from herre.grants.fakts import FaktsGrant
from fakts.grants.remote import Manifest
from typing import List, Optional
from fakts.discovery.well_known import WellKnownDiscovery
from fakts.grants.remote.device_code import DeviceCodeGrant
from fakts.grants.remote import Manifest


class ArkitektFakts(Fakts):
    pass


def build_arkitekt_fakts(
    manifest: Manifest, url: str = None, no_cache: bool = False, headless: bool = False
) -> ArkitektFakts:
    identifier = manifest.identifier
    version = manifest.version

    return ArkitektFakts(
        grant=CacheGrant(
            cache_file=f".arkitekt/cache/{identifier}-{version}_cache.json",
            hash=f"{identifier}-{version}-{url}",
            skip_cache=no_cache,
            grant=DeviceCodeGrant(
                manifest=manifest,
                open_browser=not headless,
                discovery=WellKnownDiscovery(url=url, auto_protocols=["https", "http"]),
            ),
        )
    )


def build_arkitekt_qt_fakts(
    manifest: Manifest,
    fakts: Fakts,
    url: str,
    no_cache: Optional[bool] = False,
    beacon_widget=None,
    parent=None,
):
    from fakts.grants.remote.retrieve import RetrieveGrant
    from fakts.discovery.qt.selectable_beacon import (
        SelectBeaconWidget,
        QtSelectableDiscovery,
    )

    identifier = manifest.identifier
    version = manifest.version

    beacon_widget = beacon_widget or SelectBeaconWidget(parent=parent)

    return ArkitektFakts(
        grant=CacheGrant(
            cache_file=f".arkitekt/cache/{identifier}-{version}_fakts_cache.json",
            skip_cache=no_cache,
            grant=RetrieveGrant(
                manifest=manifest,
                redirect_uri="http://localhost:6767",
                discovery=QtSelectableDiscovery(
                    widget=beacon_widget,
                    allow_appending_slash=True,
                    auto_protocols=["http", "https"],
                ),
            ),
        ),
        assert_groups={"mikro", "rekuest"},
    )



from arkitekt import register
import time


@register
def create_n_string(n: int = 10, timeout: int = 2) -> str:
    """Append World to String

    Appends world to string

    Args:
        hello (str): The first part of the string
    """
    for i in range(n):
        print(i)
        time.sleep(timeout)
        return f"Hello {i}"


@register
def append_world(hello: str) -> str:
    """Append World to String

    Appends world to string

    Args:
        hello (str): The first part of the string
    """
    return hello + " World"


@register
def print_string(input: str) -> None:
    """By World

    Prints hello world to the console

    Args:
        hello (str): The hello world string
    """
    print(input)
